@@ -1,6 +1,6 @@
-# -*- coding: iso-8859-15 -*-
+# -*- coding: utf8 -*-
 """This is just here to make pylint happy """
-# Copyright (C) 2002-2009  Tobias Klausmann
+# Copyright (C) 2002-2010  Tobias Klausmann
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -23,10 +23,10 @@ import urllib2
 
 __author__ = "klausman-pymetar@schwarzvogel.de"
 
-__version__ = "0.15"
-__revision__ = "$Rev: 84 $"[6:-2]
+__version__ = "0.17"
+__revision__ = "$Rev: 116 $"[6:-2]
 
-__doc__ = """Pymetar v%s (c) 2002-2009 Tobias Klausmann
+__doc__ = """Pymetar v%s (c) 2002-2010 Tobias Klausmann
 
 Pymetar is a python module and command line tool designed to fetch Metar
 reports from the NOAA (http://www.noaa.gov) and allow access to the
@@ -49,9 +49,9 @@ with this program; if not, write to the Free Software Foundation, Inc.,
 Please e-mail bugs to: %s""" % (__version__, __author__)
 
 
-CLOUD_RE_STR = r"^(CAVOK|CLR|SKC|BKN|SCT|FEW|OVC|NSC)([0-9]{3})?$"
-COND_RE_STR  = r"^(-|\\+)?(VC|MI|BC|PR|TS|BL|SH|DR|FZ)?\
-(DZ|RA|SN|SG|IC|PE|GR|GS|UP|BR|FG|FU|VA|SA|HZ|PY|DU|SQ|SS|DS|PO|\\+?FC)$"
+CLOUD_RE_STR = r"^(CAVOK|CLR|SKC|BKN|SCT|FEW|OVC|NSC)([0-9]{3})?(TCU|CU|CB|SC|CBMAM|ACC|SCSL|CCSL|ACSL)?$"
+CTYPE_RE_STR = r"^(TCU|CU|CB|SC|CBMAM|ACC|SCSL|CCSL|ACSL)?$"
+COND_RE_STR  = r"^(-|\\+)?(VC|MI|BC|PR|TS|BL|SH|DR|FZ)?(DZ|RA|SN|SG|IC|PE|GR|GS|UP|BR|FG|FU|VA|SA|HZ|PY|DU|SQ|SS|DS|PO|\\+?FC)$"
 
 class EmptyReportException(Exception):
     """This gets thrown when the ReportParser gets fed an empty report"""
@@ -357,25 +357,38 @@ _WeatherConditions = {
                              }),
                     }
 
+CLOUDTYPES = {
+               "ACC": "altocumulus castellanus",
+               "ACSL": "standing lenticular altocumulus",
+               "CB": "cumulonimbus",
+               "CBMAM": "cumulonimbus mammatus",
+               "CCSL": "standing lenticular cirrocumulus",
+               "CU": "cumulus",
+               "SCSL": "standing lenticular stratocumulus",
+               "SC": "stratocumulus",
+               "TCU": "towering cumulus"
+}
+
+
 def metar_to_iso8601(metardate) :
     """Convert a metar date to an ISO8601 date."""
     if metardate is not None:
         (date, hour) = metardate.split()[:2]
         (year, month, day) = date.split('.')
         # assuming tz is always 'UTC', aka 'Z'
-        return "%s-%s-%s %s:%s:00Z" % \
-            (year, month, day, hour[:2], hour[2:4])
+        return ("%s-%s-%s %s:%s:00Z" % 
+                (year, month, day, hour[:2], hour[2:4]))
 
 def parseLatLong(latlong):
     """
-    Parse Lat or Long in METAR notation into float values. N and E 
-    are +, S and W are -. Expects one positional string and returns 
+    Parse Lat or Long in METAR notation into float values. N and E
+    are +, S and W are -. Expects one positional string and returns
     one float value.
     """
     # I know, I could invert this if and put
     # the rest of the function into its block,
     # but I find it to be more readable this way
-    if latlong is None: 
+    if latlong is None:
         return None
 
     s = latlong.upper().strip()
@@ -409,7 +422,7 @@ def parseLatLong(latlong):
 
 class WeatherReport:
     """Incorporates both the unparsed textual representation of the
-    weather report and the parsed values as soon as they are filled 
+    weather report and the parsed values as soon as they are filled
     in by ReportParser."""
 
     def _ClearAllFields(self):
@@ -422,6 +435,7 @@ class WeatherReport:
         self.temp = None
         self.tempf = None
         self.windspeed = None
+        self.windspeedmph = None
         self.winddir = None
         self.vis = None
         self.dewp = None
@@ -444,6 +458,11 @@ class WeatherReport:
         self.reporturl = None
         self.latf = None
         self.longf = None
+        self.cloudinfo = None
+        self.conditions = None
+        self.w_chill = None
+        self.w_chillf = None
+        self.cloudtype = None
 
     def __init__(self, MetarStationCode = None):
         """Clear all fields and fill in wanted station id."""
@@ -504,7 +523,7 @@ class WeatherReport:
         Return the wind speed in knots
         """
         if self.windspeed is not None:
-            return self.windspeedknots
+            return self.windspeed * 1.94384449 
 
     def getWindDirection(self):
         """
@@ -584,9 +603,9 @@ class WeatherReport:
     def getCycle(self):
         """
         Return cycle value.
-        The cycle value is not the frequency or delay between 
-        observations but the "time slot" in which the observation was made. 
-        There are 24 cycle slots every day which usually last from N:45 to 
+        The cycle value is not the frequency or delay between
+        observations but the "time slot" in which the observation was made.
+        There are 24 cycle slots every day which usually last from N:45 to
         N+1:45. The cycle from 23:45 to 0:45 is cycle 0.
         """
         return self.cycle
@@ -594,19 +613,19 @@ class WeatherReport:
     def getStationPosition(self):
         """
         Return latitude, longitude and altitude above sea level of station
-        as a tuple. Some stations don't deliver altitude, for those, None 
+        as a tuple. Some stations don't deliver altitude, for those, None
         is returned as altitude.  The lat/longs are expressed as follows:
         xx-yyd
         where xx is degrees, yy minutes and d the direction.
-        Thus 51-14N means 51 degrees, 14 minutes north.  d may take the 
-        values N, S for latitues and W, E for longitudes. Latitude and 
-        Longitude may include seconds.  Altitude is always given as meters 
+        Thus 51-14N means 51 degrees, 14 minutes north.  d may take the
+        values N, S for latitues and W, E for longitudes. Latitude and
+        Longitude may include seconds.  Altitude is always given as meters
         above sea level, including a trailing M.
         Schipohl Int. Airport Amsterdam has, for example:
         ('52-18N', '004-46E', '-2M')
         Moenchengladbach (where I live):
         ('51-14N', '063-03E', None)
-        If you need lat and long as float values, look at 
+        If you need lat and long as float values, look at
         getStationPositionFloat() instead
         """
         # convert self.altitude to string for consistency
@@ -614,7 +633,7 @@ class WeatherReport:
 
     def getStationPositionFloat(self):
         """
-        Return latitude and longitude as float values in a 
+        Return latitude and longitude as float values in a
         tuple (lat, long, alt).
         """
         return (self.latf, self.longf, self.altitude)
@@ -665,7 +684,7 @@ class WeatherReport:
 
     def getTime(self):
         """
-        Return the time when the observation was made.  Note that this 
+        Return the time when the observation was made.  Note that this
         is *not* the time when the report was fetched by us
         Format:  YYYY.MM.DD HHMM UTC
         Example: 2002.04.01 1020 UTC
@@ -681,14 +700,66 @@ class WeatherReport:
 
     def getPixmap(self):
         """
-        Return a suggested pixmap name, without extension, depending on 
+        Return a suggested pixmap name, without extension, depending on
         current weather.
         """
         return self.pixmap
 
+    def getCloudinfo(self):
+        """
+        Return a tuple consisting of the parsed cloud information and a
+        suggest pixmap name
+        """
+        return self.cloudinfo
+
+    def getConditions(self):
+        """
+        Return a tuple consisting of the parsed sky conditions and a
+        suggested pixmap name
+        """
+        return self.conditions
+
+    def getWindchill(self):
+        """
+        Return wind chill in degrees Celsius
+        """
+        # http://en.wikipedia.org/wiki/Wind_chill - North American wind chill
+        # index
+        if self.w_chill is None:
+            if (self.temp and self.temp <= 10 and 
+                self.windspeed and (self.windspeed*3.6) > 4.8):
+                self.w_chill = (13.12 + 0.6215*self.temp - 
+                                11.37*(self.windspeed*3.6)**0.16 +
+                                0.3965*self.temp*(self.windspeed*3.6)**0.16)
+        return self.w_chill
+
+    def getWindchillF(self):
+        """
+        Return wind chill in degrees Fahrenheit
+        """
+        # http://en.wikipedia.org/wiki/Wind_chill - North American wind chill
+        # index
+        if self.w_chillf is None:
+            if (self.tempf and self.tempf <= 50 and 
+                self.windspeedmph and self.windspeedmph >= 3):
+                self.w_chillf = (35.74 + 0.6215*self.tempf - 
+                                 35.75*self.windspeedmph**0.16 + 
+                                 0.4275*self.tempf*self.windspeedmph**0.16)
+            else:
+                self.w_chillf = self.tempf
+
+        return self.w_chillf
+
+    def getCloudtype(self):
+        """
+        Return cloud type information
+        """
+        return self.cloudtype
+
+
 
 class ReportParser:
-    """Parse raw METAR data from a WeatherReport object into actual 
+    """Parse raw METAR data from a WeatherReport object into actual
     values and return the object with the values filled in."""
 
     def __init__(self, MetarReport = None):
@@ -698,39 +769,50 @@ class ReportParser:
     def extractCloudInformation(self) :
         """
         Extract cloud information. Return None or a tuple (sky type as a
-        string of text and suggested pixmap name)
-        """   
-        wcloud = self.match_WeatherPart(CLOUD_RE_STR)
-        if wcloud is not None :
-            stype = wcloud[:3]
-            if stype in ("CLR", "SKC", "CAV", "NSC"):
-                return ("Clear sky", "sun")
-            elif stype == "BKN" :
-                return ("Broken clouds", "suncloud")
-            elif stype == "SCT" :
-                return ("Scattered clouds", "suncloud")
-            elif stype == "FEW" :
-                return ("Few clouds", "suncloud")
-            elif stype == "OVC" :
-                return ("Overcast", "cloud")
-        else:
-            return None # Not strictly necessary
+        string of text, cloud type (if any)  and suggested pixmap name)
+        """
+        matches = self.match_WeatherPart(CLOUD_RE_STR)
+        skytype = None
+        ctype = None
+        pixmap = None
+        for wcloud in matches:
+            if wcloud is not None:
+                stype = wcloud[:3]
+                if stype in ("CLR", "SKC", "CAV", "NSC"):
+                    skytype = "Clear sky"
+                    pixmap = "sun"
+                elif stype == "BKN" :
+                    skytype = "Broken clouds"
+                    pixmap = "suncloud"
+                elif stype == "SCT" :
+                    skytype = "Scattered clouds"
+                    pixmap = "suncloud"
+                elif stype == "FEW" :
+                    skytype = "Few clouds"
+                    pixmap = "suncloud"
+                elif stype == "OVC" :
+                    skytype = "Overcast"
+                    pixmap = "cloud"
+                if ctype == None:
+                    ctype = CLOUDTYPES.get(wcloud[6:], None)
+
+        return (skytype, ctype, pixmap)
 
     def extractSkyConditions(self) :
         """
-        Extract sky condition information from the encoded report. Return 
-        a tuple containing the description of the sky conditions as a 
-        string and a suggested pixmap name for an icon representing said 
+        Extract sky condition information from the encoded report. Return
+        a tuple containing the description of the sky conditions as a
+        string and a suggested pixmap name for an icon representing said
         sky condition.
         """
-        wcond = self.match_WeatherPart(COND_RE_STR)
-        if wcond is not None :
-            if (len(wcond)>3) and (wcond.startswith('+') \
-                or wcond.startswith('-')) :
+        matches = self.match_WeatherPart(CTYPE_RE_STR)
+        for wcond in matches:
+            if ((len(wcond)>3) and 
+                (wcond.startswith('+') or wcond.startswith('-'))):
                 wcond = wcond[1:]
-            if wcond.startswith('+') or wcond.startswith('-') :
+            if wcond.startswith('+') or wcond.startswith('-'):
                 pphen = 1
-            elif len(wcond) < 4 :
+            elif len(wcond) < 4:
                 pphen = 0
             else :
                 pphen = 2
@@ -748,26 +830,28 @@ class ReportParser:
 
     def match_WeatherPart(self, regexp) :
         """
-        Return the matching part of the encoded Metar report.  
-        regexp: the regexp needed to extract this part.  
-        Return the first matching string or None.  
-        WARNING: Some Metar reports may contain several matching 
+        Return the matching part of the encoded Metar report.
+        regexp: the regexp needed to extract this part.
+        Return the first matching string or None.
+        WARNING: Some Metar reports may contain several matching
         strings, only the first one is taken into account!
         """
+        matches = []
         if self.Report.code is not None :
             rg = re.compile(regexp)
             for wpart in self.Report.getRawMetarCode().split() :
                 match = rg.match(wpart)
                 if match:
-                    return match.string[match.start(0) : match.end(0)]
+                    matches.append(match.string[match.start(0) : match.end(0)])
+        return matches
 
     def ParseReport(self, MetarReport = None):
-        """Take report with raw info only and return it with in 
+        """Take report with raw info only and return it with in
         parsed values filled in. Note: This function edits the
         WeatherReport object you supply!"""
         if self.Report is None and MetarReport is None:
-            raise EmptyReportException, \
-                "No report given on init and ParseReport()."
+            raise EmptyReportException(
+                "No report given on init and ParseReport().")
         elif MetarReport is not None:
             self.Report = MetarReport
 
@@ -803,6 +887,12 @@ class ReportParser:
                 except ValueError:
                     (lat, lng) = p.split()[1:3]
                     ht = None
+                # A few jokers out there think O==0
+                if "O" in lat:
+                    lat = lat.replace("O", "0")
+                if "O" in lng:
+                    lng = lng.replace("O", "0")
+
                 self.Report.stat_city = rcity.strip()[::-1]
                 self.Report.stat_country = rcoun.strip()[::-1]
                 self.Report.fulln = loc
@@ -815,8 +905,8 @@ class ReportParser:
             # The line containing date and time of the report
             # We have to make sure that the station ID is *not*
             # in this line to avoid trying to parse the ob: line
-            elif ((data.find("UTC")) != -1 and \
-                    (data.find(self.Report.givenstationid)) == -1):
+            elif ((data.find(" UTC")) != -1 and 
+                  (data.find(self.Report.givenstationid)) == -1):
                 rt = data.split("/")[1]
                 self.Report.rtime = rt.strip()
 
@@ -828,9 +918,16 @@ class ReportParser:
                 # The string we have split is "(NN C)", hence the slice
                 self.Report.temp = float(c[1:])
 
+            # wind chill
+
+            elif (header == "Windchill"):
+                f, c = data.split(None, 3)[0:3:2]
+                self.Report.w_chillf = float(f)
+                # The string we have split is "(NN C)", hence the slice
+                self.Report.w_chill = float(c[1:])
 
             # wind dir and speed
-            
+
             elif (header == "Wind"):
                 if (data.find("Calm") != -1):
                     self.Report.windspeed = 0.0
@@ -871,7 +968,7 @@ class ReportParser:
                         break
 
             # dew point
-            
+
             elif (header == "Dew Point"):
                 f, c = data.split(None, 3)[0:3:2]
                 self.Report.dewpf = float(f)
@@ -879,19 +976,19 @@ class ReportParser:
                 self.Report.dewp = float(c[1:])
 
             # humidity
-             
+
             elif (header == "Relative Humidity"):
                 h = data.split("%", 1)[0]
                 self.Report.humid = int(h)
 
             # pressure
-            
+
             elif (header == "Pressure (altimeter)"):
                 p = data.split(" ", 1)[0]
                 self.Report.press = (float(p)*33.863886)
 
             # shot weather desc. ("rain", "mist", ...)
-            
+
             elif (header == "Weather"):
                 self.Report.weather = data
 
@@ -901,27 +998,35 @@ class ReportParser:
                 self.Report.sky = data
 
             # the encoded report itself
-            
+
             elif (header == "ob"):
                 self.Report.code = data.strip()
 
             # the cycle value ("time slot")
 
             elif (header == "cycle"):
-                self.Report.cycle = int(data)
+                try:
+                    self.Report.cycle = int(data)
+                except ValueError:
+                    # cycle value is missing or garbled, assume cycle 0
+                    # TODO: parse the date/time header if it isn't too involved
+                    self.Report.cycle = 0
+
 
         # cloud info
         cloudinfo = self.extractCloudInformation()
-        if cloudinfo is not None :
-            (cloudinfo, cloudpixmap) = cloudinfo
-        else :
-            (cloudinfo, cloudpixmap) = (None, None)
+        (cloudinfo, cloudtype, cloudpixmap) = cloudinfo
+
         conditions = self.extractSkyConditions()
         if conditions is not None :
             (conditions, condpixmap) = conditions
         else :
             (conditions, condpixmap) = (None, None)
 
+        # Some people might want to always use sky or cloud info specifially
+        self.Report.cloudinfo = (cloudinfo, cloudpixmap)
+        self.Report.conditions = (conditions, condpixmap)
+
         # fill the weather information
         self.Report.weather = self.Report.weather or conditions or cloudinfo
 
@@ -929,6 +1034,10 @@ class ReportParser:
         # over pixmap guessed from clouds
         self.Report.pixmap = condpixmap or cloudpixmap
 
+        # Cloud type (Cumulonimbus etc.)
+        if self.Report.cloudtype == None:
+            self.Report.cloudtype = cloudtype
+
         # report is complete
         self.Report.valid = 1
 
@@ -936,10 +1045,10 @@ class ReportParser:
 
 class ReportFetcher:
     """Fetches a report from a given METAR id, optionally taking into
-       account a different baseurl and using environment var-specified 
+       account a different baseurl and using environment var-specified
        proxies."""
 
-    def __init__(self, MetarStationCode = None, baseurl = \
+    def __init__(self, MetarStationCode = None, baseurl = 
         "http://weather.noaa.gov/pub/data/observations/metar/decoded/"):
         """Set stationid attribute and base URL to fetch report from"""
         self.stationid = MetarStationCode
@@ -967,18 +1076,18 @@ class ReportFetcher:
         protocol://user:password@host.name.tld:port/
         is expected, for example:
         http://squid.somenet.com:3128/
-        If no proxy is specified, the environment variable http_proxy 
+        If no proxy is specified, the environment variable http_proxy
         is inspected. If it isn't set, a direct connection is tried.
         """
         if self.stationid is None and StationCode is None:
-            raise EmptyIDException, \
-                "No ID given on init and FetchReport()."
+            raise EmptyIDException( 
+                "No ID given on init and FetchReport().")
         elif StationCode is not None:
             self.stationid = StationCode
 
         self.stationid = self.stationid.upper()
         self.reporturl = "%s%s.TXT" % (self.baseurl, self.stationid)
-        
+
         if proxy:
             p_dict = {'http': proxy}
             p_handler = urllib2.ProxyHandler(p_dict)
@@ -998,14 +1107,14 @@ class ReportFetcher:
 
         if fn.info().status:
             raise NetworkException, "Could not fetch METAR report"
-            
+
         report = WeatherReport(self.stationid)
         report.reporturl = self.reporturl
         report.fullreport = self.fullreport
         self.report = report # Caching it for GetReport()
 
         return report
-        
+
     def GetReport(self):
         """Get a previously fetched report again"""
         return self.report
